
/*****************************************************************************
*  COPYRIGHT : This code is written by, and copyright to Graeme F Prentice.
*              You may not distribute this code in whole or in part or use it
*              in a commercial product without permission from the author in
*              writing.   
*****************************************************************************/


/******************************************************************************
*  $Revision: 1.1 $                                                            
******************************************************************************/


#include "slick.sh"
#include "Listman.sh"

#pragma option(strictsemicolons,on)
#pragma option(strict,on)
#pragma option(autodecl,off)
#pragma option(strictparens,on)

defeventtab ListmanConfigForm;

// don't need this right now
#define LISTMAN_CONFIG_VERSION 'V1_00'

_control top_separator_image;
_control listnames_fontname_editbox;
_control listnames_font_size_editbox;


_control filenames_fontname_editbox;
_control filenames_font_size_editbox;
_control show_filenames_col_header_checkbox;
_control width_col0_textbox;
_control width_col1_textbox;

_control AllowResizeTreeCol0_checkbox;

Listman_persistent_setup_data Listman_global_persistent_data;




#define CFG_FUNC_SET_DEFAULT 1
#define CFG_FUNC_SET_ITV_TO_PPV 2
#define CFG_FUNC_INSERT_ITEM 3
#define CFG_FUNC_SET_ITV_TO_VAL 4
#define CFG_FUNC_SET_PPV_TO_ITV 5

// in MAKE_CONFIG_FUNC macro, item_name becomes the name of a function used to
// perform operations for this item and is also the name of the item in the
// ini file and the member name of persistent_config_data struct.  
// For the CFG_FUNC_SET_ITV_TO_VAL operation, only the item whose
// name matches is affected.  ptype is used to typecast a value before it is
// assigned to the item.
#define MAKE_CONFIG_FUNC(item_name, property_name, default_val, ptype)    \
static void item_name(int what, Listman_persistent_config_data * config_ptr, typeless val, _str field_name)\
{\
   switch (what) {\
      case CFG_FUNC_SET_DEFAULT : config_ptr->##item_name = (ptype)default_val; return;\
      case CFG_FUNC_SET_ITV_TO_PPV : config_ptr->##item_name = (ptype)property_name; return;\
      case CFG_FUNC_INSERT_ITEM : insert_line( #item_name '=' :+ config_ptr->##item_name); return;\
      case CFG_FUNC_SET_ITV_TO_VAL : \
         if (strcmp(field_name,#item_name) == 0) config_ptr->##item_name = (ptype)val;\
         return;\
      case CFG_FUNC_SET_PPV_TO_ITV : property_name = config_ptr->##item_name; return;\
      default : return;\
   }\
}


   // generate all the config functions
   MAKE_CONFIG_FUNC(listnames_fontname,              p_active_form.listnames_fontname_editbox.p_text, 'Microsoft Sans Serif', _str)
   MAKE_CONFIG_FUNC(listnames_font_size,         p_active_form.listnames_font_size_editbox.p_text, 8, _str)

   MAKE_CONFIG_FUNC(filenames_fontname,   p_active_form.filenames_fontname_editbox.p_text, 'Microsoft Sans Serif', _str)
   MAKE_CONFIG_FUNC(filenames_font_size, p_active_form.filenames_font_size_editbox.p_text, 8, _str)
   MAKE_CONFIG_FUNC(show_filenames_listbox_column_header, p_active_form.show_filenames_col_header_checkbox.p_value, 1, int)
   MAKE_CONFIG_FUNC(width_col0, p_active_form.width_col0_textbox.p_text, '10', _str)
   MAKE_CONFIG_FUNC(width_col1, p_active_form.width_col1_textbox.p_text, '22', _str)
   MAKE_CONFIG_FUNC(allow_resize_tree_col0, p_active_form.AllowResizeTreeCol0_checkbox.p_value, 0, int)

   // items added to this list should also be added to call_config_funcs below


#define CALL(item_name,val) item_name(what, config_ptr, val, field_name)

static call_config_funcs(int what, Listman_persistent_config_data * config_ptr, typeless val = 0, _str field_name = '')
{
   CALL(listnames_fontname, val);
   CALL(listnames_font_size, val);
   CALL(filenames_fontname, val);
   CALL(filenames_font_size, val);
   CALL(show_filenames_listbox_column_header, val);
   CALL(width_col0, val);
   CALL(width_col1, val);
   CALL(allow_resize_tree_col0, val);
}

#undef CALL


static void Listman_save_config( Listman_persistent_setup_data * dptr, _str section_name, boolean no_form = false )
{
   // copy widget property values into item values
   if (!no_form)
      call_config_funcs(CFG_FUNC_SET_ITV_TO_PPV, &dptr->config);

   int section_view, current_view;
   current_view = _create_temp_view(section_view);
   if (current_view == '') {
       return;
   }
   activate_view(section_view);
   top();

   insert_line('Listman_config_version=' :+ LISTMAN_CONFIG_VERSION);
   // write all item values to the active view
   call_config_funcs(CFG_FUNC_INSERT_ITEM, &dptr->config, 0);

   int res = _ini_replace_section(_config_path() :+ 'ListmanConfig.ini', section_name, section_view);
}


void Listman_load_config( Listman_persistent_setup_data * dptr, _str section_name)
{
    Listman_persistent_config_data * config_ptr = &dptr->config;

    // set default values for all items
    call_config_funcs(CFG_FUNC_SET_DEFAULT, &dptr->config);

    int section_view, current_view;
    get_view_id(current_view);
    int res;
    res = _ini_get_section(_config_path() :+ 'ListmanConfig.ini', section_name, section_view);
    if (res) {
        return;
    }
    activate_view(section_view);
    top();
    _insert_text(' '\r\n);  // _ini_parse_line does a down() first
    top();
    activate_view(current_view);
    _str field_name,value;
    int k = 0, maxl = 0;
    while ( k==0 ) {
        if (++maxl > 4000) {
            break;
        }
        k = _ini_parse_line(section_view,field_name,value);
        if (k==0) {
           // Set item value whose name matches the field name, to the parsed value
           // This was previously a switch statement with case options being the
           // item names - which might be faster if it uses binary search style match.
           call_config_funcs(CFG_FUNC_SET_ITV_TO_VAL, config_ptr, value, field_name);
        }
    }
}


static void populate_config_dialog_options( Listman_persistent_setup_data * dptr )
{
   call_config_funcs(CFG_FUNC_SET_PPV_TO_ITV, &dptr->config);
}



void listnames_font_change_button.lbutton_up()
{
   _str font = _font_param( p_active_form.listnames_fontname_editbox.p_text,
                      p_active_form.listnames_font_size_editbox.p_text, 0);
   int result = show('-modal _font_form',
               '',      // Display fixed fonts only
                font
              );
   if (result == '') {
    return;
   }
   _str font_name, font_size, font_flags;
   parse result with font_name','font_size','font_flags',';
   p_active_form.listnames_fontname_editbox.p_text = font_name;
   p_active_form.listnames_font_size_editbox.p_text = font_size;
}


void filenames_font_change_button.lbutton_up()
{
    _str font = _font_param( p_active_form.filenames_fontname_editbox.p_text,
                        p_active_form.filenames_font_size_editbox.p_text, 0);
   int result = show('-modal _font_form',
                 '',      // Display fixed fonts only
                  font
                );
   if (result == '') {
      return;
   }
   _str font_name, font_size, font_flags;
   parse result with font_name','font_size','font_flags',';
   p_active_form.filenames_fontname_editbox.p_text = font_name;
   p_active_form.filenames_font_size_editbox.p_text = font_size;
}



void ListmanConfigForm.on_create(_str arg1 = '')
{
   // todo - use macro for separator
    int pic1 = _update_picture(0,GFM_BITMAP_PATH :+ '/GFM_separator1.bmp');

    top_separator_image.p_picture = pic1;
    top_separator_image.p_stretch = true;
    top_separator_image.p_auto_size = true;
    top_separator_image.p_x = 0;
    //top_separator_image.p_y = pix2scale(50,p_active_form);

    top_separator_image.p_height = pix2scale(2,p_active_form);         
    top_separator_image.p_width =
        pix2scale(p_active_form.p_client_width,p_active_form);

    Listman_load_config(&Listman_global_persistent_data, 'ListmanConfig' );

    call_config_funcs(CFG_FUNC_SET_PPV_TO_ITV, &Listman_global_persistent_data.config);
}


void ListmanSaveColWidths(int wcol0, int wcol1)
{
   Listman_load_config(&Listman_global_persistent_data, 'ListmanConfig' );
   Listman_global_persistent_data.config.width_col0 = wcol0;
   Listman_global_persistent_data.config.width_col1 = wcol1;
   Listman_save_config(&Listman_global_persistent_data, 'ListmanConfig', true );
}


void save_button.lbutton_up()
{
    Listman_save_config(&Listman_global_persistent_data, 'ListmanConfig' );
}


void ok_button.lbutton_up()
{
   Listman_save_config(&Listman_global_persistent_data, 'ListmanConfig' );

   int form_id = _find_object('ListmanConfigForm');
   if (form_id != 0) 
      form_id._delete_window('');
   ListmanProcessNewConfig();
}


definit()
{
   Listman_load_config(&Listman_global_persistent_data, 'ListmanConfig' );
}




